/***************************************************************************
 *   Copyright (C) 2004 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#include "bezierline.h"

namespace Papyrus {

Bezierline::Bezierline(BezierVertices vertices):
    Shape(),
    m_vertices(vertices) {
      set_use_outline( true );
    }

    Bezierline::pointer Bezierline::create(BezierVertices vertices) {
        return Bezierline::pointer( new Bezierline( vertices ) );
    }

    Bezierline::~Bezierline() {}

    Bezierline & Bezierline::operator =( const Bezierline & other )
    {
      Shape::operator=( other );
      m_vertices = other.m_vertices;

      if (other.m_start_marker)
        m_start_marker = other.m_start_marker->clone();
      else
        m_start_marker = other.m_start_marker;

      if (other.m_end_marker)
        m_end_marker = other.m_end_marker->clone();
      else
        m_end_marker = other.m_end_marker;

      if (other.m_vertex_marker)
        m_vertex_marker = other.m_vertex_marker->clone();
      else
        m_vertex_marker = other.m_vertex_marker;

      m_signal_vertices.emit();
      m_signal_start_marker.emit();
      m_signal_end_marker.emit();
      m_signal_vertex_marker.emit();
    }

    BezierVertex Bezierline::get_vertex(unsigned n) {
			if (n < m_vertices.size())
				return m_vertices[n];
			else
				return BezierVertex();
		}

    const BezierVertices& Bezierline::get_vertices() {
			return m_vertices;
		}

		void Bezierline::add_vertex(BezierVertex v) {
			m_vertices.push_back(v);
			this->on_vertex_changed(m_vertices.size()-1);
		}

    void Bezierline::add_vertex(double x, double y, double c1x, double c1y, double c2x, double c2y) {
      m_vertices.push_back(BezierVertex(x,y, c1x, c1y, c2x, c2y));
			this->on_vertex_changed(m_vertices.size()-1);
		}

		void Bezierline::set_vertex(unsigned n, BezierVertex v) {
			if (n < m_vertices.size()) {
				m_vertices[n] = v;
				this->on_vertex_changed(n);
			}
		}

    void Bezierline::set_vertex(unsigned n, double x, double y, double c1x, double c1y, double c2x, double c2y) {
			if (n < m_vertices.size()) {
				m_vertices[n] = BezierVertex(x, y, c1x, c1y, c2x, c2y);
				this->on_vertex_changed(n);
			}
		}

		void Bezierline::set_vertices(BezierVertices vertices) {
			m_vertices = vertices;
			this->on_vertices_changed();
		}

    LineStyle& Bezierline::line_style() {
        return outline();
    }

    void Bezierline::set_line_style(const LineStyle& style) {
        set_outline( style, true );
    }

    Drawable::pointer Bezierline::start_marker( )
    {
      return m_start_marker;
    }

    void Bezierline::set_start_marker( Drawable::pointer marker )
    {
      if ( marker )
        m_start_marker = marker->clone();
      else
        m_start_marker = marker;

			this->set_extents( m_vertex_extents );
			m_signal_start_marker.emit();
      m_signal_changed.emit();
    }

    Drawable::pointer Bezierline::end_marker( )
    {
      return m_end_marker;
    }

    void Bezierline::set_end_marker( Drawable::pointer marker )
    {
      if ( marker )
        m_end_marker = marker->clone();
      else
        m_end_marker = marker;

			this->set_extents( m_vertex_extents );
			m_signal_end_marker.emit();
      m_signal_changed.emit();
    }

    Drawable::pointer Bezierline::vertex_marker( )
    {
      return m_vertex_marker;
    }

    void Bezierline::set_vertex_marker( Drawable::pointer marker )
    {
      if ( marker )
        m_vertex_marker = marker->clone();
      else
        m_vertex_marker = marker;

      m_signal_vertex_marker.emit();
      m_signal_changed.emit();
    }

    void Bezierline::draw(Cairo::RefPtr<Cairo::Context> cairo) {
      if (m_fill) {
        this->draw_shape(cairo);
        cairo->set_source(m_fill);
        cairo->fill( );
      }

      if (m_use_outline) {
        m_outline.render( cairo );
        this->draw_outline(cairo);
        cairo->stroke( );
      }

      if (m_vertices.size() < 1)
        return;

      BezierVertices::iterator i1, i2;
      BezierVertices::reverse_iterator ri1, ri2;

      if ( m_start_marker ) {
        i1 = m_vertices.begin();
        i2 = m_vertices.begin() + 1;
        m_start_marker->set_translate( i1->x, i1->y );
        if ( i2 != m_vertices.end() )
          m_start_marker->set_rotate(
              atan ( ( i2->y - i1->y ) /
              ( i2->x - i1->x )
                   ) );
        m_start_marker->render( cairo );
      }

      if ( m_vertex_marker ) {
      }

      if ( m_end_marker ) {
        ri1 = m_vertices.rbegin();
        ri2 = m_vertices.rbegin() + 1;
        m_end_marker->set_translate( ri1->x, ri1->y );
        if ( ri2 != m_vertices.rend() )
          m_end_marker->set_rotate(
              atan ( ( ri2->y - ri1->y ) /
              ( ri2->x - ri1->x )
                   ) );
        m_end_marker->render( cairo );
      }

    }

    void Bezierline::draw_shape( Cairo::RefPtr<Cairo::Context> cairo) {
        if (m_vertices.size() < 2)
            return;
        cairo->move_to( m_vertices[0].x, m_vertices[0].y);
        BezierVertices::iterator prev=m_vertices.begin(), current=m_vertices.begin()+1;
        while ( current != m_vertices.end() )
        {
          if ( current->relative )
            cairo->rel_curve_to( prev->c2x, prev->c2y, current->c1x, current->c1y, current->x, current->y );
          else
            cairo->curve_to( prev->c2x, prev->c2y, current->c1x, current->c1y, current->x, current->y );
          prev = current++;
        }
    }

    sigc::signal<void, unsigned>& Bezierline::signal_vertex() {
      return m_signal_vertex;
    }

    sigc::signal<void>& Bezierline::signal_vertices() {
      return m_signal_vertices;
    }

    sigc::signal<void> Bezierline::signal_start_marker() {
      return m_signal_start_marker;
    }

    sigc::signal<void> Bezierline::signal_end_marker() {
      return m_signal_end_marker;
    }

    sigc::signal<void> Bezierline::signal_vertex_marker() {
      return m_signal_vertex_marker;
    }

    void Bezierline::on_vertex_changed(int n) {
			m_vertex_extents.include( m_vertices[n].x, m_vertices[n].y );
			this->set_extents( m_vertex_extents );
			this->shape_changed();
			m_signal_vertex.emit(n);
			m_signal_vertices.emit();
		}

		void Bezierline::on_vertices_changed() {
      if (m_vertices.size() > 0 ) {
        m_vertex_extents.set_xywh( m_vertices[0].x, m_vertices[0].y, 0.0, 0.0 );
        for (BezierVertices::iterator i = m_vertices.begin()+1; i != m_vertices.end(); i++)
          m_vertex_extents.include(i->x, i->y);
      }
      else
        m_vertex_extents.set_xywh( 0.0, 0.0, 0.0, 0.0 );

      this->set_extents( m_vertex_extents );
      this->shape_changed();
			m_signal_vertices.emit();
		}

		void Bezierline::set_extents(const Region& extents) {
			if ( m_start_marker || m_end_marker ) {
				Region updated_extents = extents;
				if ( m_start_marker )
					updated_extents.include( m_start_marker->extents() );
				if ( m_end_marker )
					updated_extents.include( m_end_marker->extents() );

				this->Shape::set_extents( updated_extents );
			}
			else
				this->Shape::set_extents( extents );
		}

}
