/***************************************************************************
 *   Copyright (C) 2004 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/

#include "factory.h"

#include "example_image.h"

namespace Papyrus {

char* DrawableStrings[] = {
    "Arc",
    "Bezierline",
    "Beziergon",
    "Boxed",
    "Circle",
    "Face",
    "Grid",
    "Group",
    "Handle Box",
    "Hilbert Curve",
    "Image",
    "Koch Curve",
    "Koch Snowflake",
    "Polygon",
    "Polyline",
    "Rectangle",
    "Regular Polygon",
    "Sierpinski",
    "Text",
};

Arc::pointer example_arc(bool fill, bool outline) {
  // Create an arc with a radius of 50 that will start at 45' and stop at 315'
  Papyrus::Arc::pointer arc = Papyrus::Arc::create(50, M_PI_4, 3.0*M_PI_2+M_PI_4);

  // Set the fill color with an alpha value of 0.9
  if (fill)
    arc->set_fill( Cairo::SolidPattern::create_rgba(1.0, 1.0, 0.0, 0.9) );

  // And set the outline color
  if (outline)
    arc->set_outline( Cairo::SolidPattern::create_rgba(0.0, 0.0, 0.0, 0.9) );

  return arc;
}

Bezierline::pointer example_bezierline(bool fill, bool outline)
{
  Bezierline::pointer shape = Bezierline::create();
  double phi = 1.6180333989;
  double four_spline_magic_number = 0.55228475;
  double side = 100.0 / ( 1.0 + 1.0/phi );
  double yoffset = side / 2.0;
  double x0=0.0, y0=0.0, x1, y1, c0x=0.0, c0y=0.0, c1x=0.0, c1y=0.0, c2x=0.0, c2y=0.0;
  for ( int i=0; i < 10; i++ ) {
    switch(i%4) {
    case 0:
      x1 = x0 + side;
      y1 = y0 - side;
      c1x = x0;
      c1y = y0 - side * four_spline_magic_number;
      c2x = x1 - side * four_spline_magic_number;
      c2y = y1;
      break;
    case 1:
      x1 = x0 + side;
      y1 = y0 + side;
      c1x = x0 + side * four_spline_magic_number;
      c1y = y0;
      c2x = x1;
      c2y = y1 - side * four_spline_magic_number;
      break;
    case 2:
      x1 = x0 - side;
      y1 = y0 + side;
      c1x = x0;
      c1y = y0 + side * four_spline_magic_number;
      c2x = x1 + side * four_spline_magic_number;
      c2y = y1;
      break;
    case 3:
      x1 = x0 - side;
      y1 = y0 - side;
      c1x = x0 - side * four_spline_magic_number;
      c1y = y0;
      c2x = x1;
      c2y = y1 + side * four_spline_magic_number;
      break;
    }
    shape->add_vertex(x0, y0, c0x, c0y, c1x, c1y);
    x0 = x1;
    y0 = y1;
    c0x = c2x;
    c0y = c2y;
    side /= phi;
  }
  shape->add_vertex(x0, y0, c0x, c0y, c1x, c1y);
  shape->set_centroid( 50.0, -yoffset );
  shape->translate( -50.0, yoffset );
  shape->scale( 1.8, 1.8 );
  if (fill)
    shape->set_fill( Cairo::SolidPattern::create_rgba( 0.8, 0.8, 1.0, 0.9 ) );
  if (outline) {
    shape->set_outline( Cairo::SolidPattern::create_rgb( 1.0, 0.0, 1.0 ) );
    shape->outline().set_width( 2.0 );
  }
  return shape;
}

Beziergon::pointer example_beziergon(bool fill, bool outline)
{
  Beziergon::pointer shape = Beziergon::create();
  double phi = 1.6180333989;
  double four_spline_magic_number = 0.55228475;
  double side = 100.0 / ( 1.0 + 1.0/phi );
  double yoffset = side / 2.0;
  double x0=0.0, y0=0.0, x1, y1, c0x=0.0, c0y=0.0, c1x=0.0, c1y=0.0, c2x=0.0, c2y=0.0;
  for ( int i=0; i < 10; i++ ) {
    switch(i%4) {
    case 0:
      x1 = x0 + side;
      y1 = y0 - side;
      c1x = x0;
      c1y = y0 - side * four_spline_magic_number;
      c2x = x1 - side * four_spline_magic_number;
      c2y = y1;
      break;
    case 1:
      x1 = x0 + side;
      y1 = y0 + side;
      c1x = x0 + side * four_spline_magic_number;
      c1y = y0;
      c2x = x1;
      c2y = y1 - side * four_spline_magic_number;
      break;
    case 2:
      x1 = x0 - side;
      y1 = y0 + side;
      c1x = x0;
      c1y = y0 + side * four_spline_magic_number;
      c2x = x1 + side * four_spline_magic_number;
      c2y = y1;
      break;
    case 3:
      x1 = x0 - side;
      y1 = y0 - side;
      c1x = x0 - side * four_spline_magic_number;
      c1y = y0;
      c2x = x1;
      c2y = y1 + side * four_spline_magic_number;
      break;
    }
    shape->add_vertex(x0, y0, c0x, c0y, c1x, c1y);
    x0 = x1;
    y0 = y1;
    c0x = c2x;
    c0y = c2y;
    side /= phi;
  }
  shape->add_vertex(x0, y0, c0x, c0y, c1x, c1y);
  shape->translate( -50.0*1.8, yoffset*1.8 );
  shape->scale( 1.8, 1.8 );
  if (fill)
    shape->set_fill( Cairo::SolidPattern::create_rgba( 1.0, 1.0, 0.0, 0.9 ) );
  if (outline) {
    shape->set_outline( Cairo::SolidPattern::create_rgb( 0.0, 1.0, 0.0 ) );
    shape->outline().set_width(2.0);
  }
  return shape;
}

Circle::pointer example_circle(bool fill, bool outline) {
  // Create a circle with a radius of 50
  Papyrus::Circle::pointer circle = Papyrus::Circle::create( 50 );

  // Set the fill color to green, with an alpha value of 0.9
  if (fill) circle->set_fill( Cairo::SolidPattern::create_rgba(0.0, 1.0, 0.0, 0.9) );

  // And the outline color to black
  if (outline) circle->set_outline( Cairo::SolidPattern::create_rgba(0.0, 0.0, 0.0, 0.9) );

  return circle;
}

Rectangle::pointer example_rectangle(bool fill, bool outline) {
  // Create a rectangle
  Papyrus::Rectangle::pointer rectangle = Papyrus::Rectangle::create( 80, 50 );

  // Set the fill color to red, with an alpha value of 0.9
  if (fill) rectangle->set_fill( Cairo::SolidPattern::create_rgba(1.0, 0.0, 0.0, 0.9) );

  // And the outline color to black
  if (outline) rectangle->set_outline( Cairo::SolidPattern::create_rgba(0.0, 0.0, 0.0, 0.9) );

  return rectangle;
}

Group::pointer example_group(bool fill, bool outline) {
  // Create a group to contain several shapes
  Papyrus::Group::pointer group = Papyrus::Group::create( );

  // Create some shapes to add to the group
  Papyrus::Rectangle::pointer rectangle = example_rectangle( fill, outline );
  Papyrus::Circle::pointer circle = example_circle( fill, outline );
  Papyrus::Arc::pointer arc = example_arc( fill, outline );

  // Add the shapes to the group
  group->add( rectangle );
  group->add( circle );
  group->add( arc );

  // Translate the shapes so they don't necessarily overlap
  rectangle->set_xywh(0, -60, 50, 30);
  circle->set_translate(40, 20);
  circle->set_radius(25);
  arc->set_translate(0, 20);

  // Rotate the entire group by 45 degrees counter-clockwise
  group->rotate(-M_PI_4);

  return group;
}

Handlebox::pointer example_handlebox(bool fill, bool outline) {
  // Create a handlebox to contain several shapes
  Papyrus::Handlebox::pointer handlebox = Papyrus::Handlebox::create( );
  if (fill) handlebox->get_box()->set_fill( Cairo::SolidPattern::create_rgba(1.0, 0.85, 1.0, 0.9) );
  if (outline) handlebox->get_box()->set_outline( Cairo::SolidPattern::create_rgba(1.0, 0.0, 1.0, 0.9) );

  // Create some shapes to add to the handlebox
  Papyrus::Rectangle::pointer rectangle = example_rectangle( );
  Papyrus::Circle::pointer circle = example_circle( );
  Papyrus::Arc::pointer arc = example_arc( );

  // Add the shapes to the handlebox
  handlebox->add( rectangle );
  handlebox->add( circle );
  handlebox->add( arc );

  // Translate the shapes so they don't necessarily overlap
  rectangle->set_xywh(0, -60, 50, 30);
  circle->set_translate(40, 20);
  circle->set_radius(25);
  arc->set_translate(0, 20);

  return handlebox;
}

Boxed::pointer example_boxed(bool fill, bool outline) {
  // Create a boxed group to contain several shapes
  Papyrus::Boxed::pointer boxed = Papyrus::Boxed::create( );
  if (fill) boxed->get_box()->set_fill( Cairo::SolidPattern::create_rgba(1.0, 0.85, 1.0, 0.9) );
  if (outline) boxed->get_box()->set_outline( Cairo::SolidPattern::create_rgba(1.0, 0.0, 1.0, 0.9) );

  // Create some shapes to add to the boxed group
  Papyrus::Rectangle::pointer rectangle = example_rectangle( );
  Papyrus::Circle::pointer circle = example_circle( );
  Papyrus::Arc::pointer arc = example_arc( );

  // Add the shapes to the boxed group
  boxed->add( rectangle );
  boxed->add( circle );
  boxed->add( arc );

  // Translate the shapes and adjust size so they don't necessarily overlap
  rectangle->set_xywh(0, -60, 50, 30);
  circle->set_translate(40, 20);
  circle->set_radius(25);
  arc->set_translate(0, 20);

  return boxed;
}

Image::pointer example_image( bool, bool )
{
  static bool convert = true;
  Image::pointer image;

  image = Image::create( papyrus_example_image.pixel_data,
                         Cairo::FORMAT_RGB24,
                         papyrus_example_image.width,
                         papyrus_example_image.height,
                         papyrus_example_image.width * papyrus_example_image.bytes_per_pixel,
                         convert
                       );
  convert = false;

  return image;
}

Drawable::pointer example_factory( unsigned int d, bool fill, bool outline )
{
  if (d >= DRAWABLE_END)
    throw;

  switch (d) {
  case ARC:
    return example_arc( fill, outline );
  case BEZIERLINE:
    return example_bezierline( fill, outline );
  case BEZIERGON:
    return example_beziergon( fill, outline );
  case BOXED:
    return example_boxed( fill, outline );
  case CIRCLE:
    {
      Circle::pointer shape = Circle::create(20);
      if (fill)
        shape->set_fill( RGBA("green", 0.9) );
      if (outline)
        shape->set_outline( Cairo::SolidPattern::create_rgb( 1.0, 0.0, 1.0 ) );
      return shape;
    }
  case FACE:
    return Face::create(20);
  case GRID:
    {
      Grid::pointer shape = Grid::create( -175, 175, -175, 175 );
      shape->set_x_interval(50.0);
      shape->set_y_interval(50.0);
      shape->set_x_interval_style( LineStyle(RGBA(0.8, 0.8, 0.8, 0.9), 1) );
      shape->set_y_interval_style( LineStyle(RGBA(0.8, 0.8, 0.8, 0.9), 1) );
      if (outline)
      {
        shape->set_x_axis_style( LineStyle(RGBA("DarkGreen", 0.9), 5) );
        shape->set_y_axis_style( LineStyle(RGBA("DarkBlue", 0.9), 5) );
        shape->enable_border();
        shape->set_border_style( LineStyle(RGBA("DarkRed", 0.9), 5) );
      }
      return shape;
    }
  case GROUP:
    return example_group( fill, outline );
  case HANDLEBOX:
    return example_handlebox( fill, outline );
  case HILBERT:
    {
      Hilbert::pointer shape = Hilbert::create( 3, 100 );
      if (fill)
        shape->set_fill( RGBA("LimeGreen", 0.9) );
      if (outline)
        shape->set_outline( RGBA("IndianRed", 0.9) );
      return shape;
    }
  case IMAGE:
    return example_image( fill, outline );
  case KOCHCURVE:
    {
      KochCurve::pointer shape = KochCurve::create( 3, 100 );
      if (fill)
        shape->set_fill( RGBA("PowderBlue", 0.9) );
      if (outline)
        shape->set_outline( RGBA("MediumBlue", 0.9) );
      return shape;
    }
  case KOCHSNOWFLAKE:
    {
      KochSnowflake::pointer shape = KochSnowflake::create( 4, 100 );
      if (fill)
        shape->set_fill( RGBA("PowderBlue", 0.9) );
      if (outline)
        shape->set_outline( RGBA("MediumBlue", 0.9) );
      return shape;
    }
  case POLYGON:
    {
      Polygon::pointer shape = Polygon::create();
      shape->add_vertex(0, 0);
      shape->add_vertex(20, 0);
      shape->add_vertex(0, 20);
      shape->add_vertex(20, 20);
      if (fill)
        shape->set_fill( Cairo::SolidPattern::create_rgb( 0.0, 1.0, 1.0 ) );
      if (outline)
        shape->set_outline( Cairo::SolidPattern::create_rgb( 1.0, 0.0, 0.0 ) );
      return shape;
    }
  case POLYLINE:
    {
      Polyline::pointer shape = Polyline::create();
      shape->add_vertex(0, 0);
      shape->add_vertex(20, 0);
      shape->add_vertex(0, 20);
      shape->add_vertex(20, 20);
      if (fill)
        shape->set_fill( Cairo::SolidPattern::create_rgb( 1.0, 0.0, 0.0 ) );
      if (outline)
        shape->set_outline( Cairo::SolidPattern::create_rgb( 0.0, 1.0, 1.0 ) );
      return shape;
    }
  case RECTANGLE:
    return example_rectangle( fill, outline );
  case REGULARPOLYGON:
    {
      RegularPolygon::pointer shape = RegularPolygon::create(8, 40, RegularPolygon::ZERO_APOTHEM);
      if (fill)
        shape->set_fill( Cairo::SolidPattern::create_rgb( 1.0, 1.0, 0.0 ) );
      if (outline)
        shape->set_outline( Cairo::SolidPattern::create_rgb( 0.0, 0.0, 1.0 ) );
      return shape;
    }
  case SIERPINSKI:
    {
      Sierpinski::pointer shape = Sierpinski::create( 3, 100 );
      if (fill)
        shape->set_fill( RGBA("LimeGreen", 0.9) );
      if (outline)
        shape->set_outline( RGBA("IndianRed", 0.9) );
      return shape;
    }
  case TEXT:
    {
      Text::pointer shape = Text::create("Text", 40);
      if (fill)
        shape->set_fill( Cairo::SolidPattern::create_rgb( 1.0, 1.0, 1.0 ) );
      if (outline)
        shape->set_outline( Cairo::SolidPattern::create_rgb( 0.0, 0.0, 0.0 ) );
      return shape;
    }
  }
  return Drawable::pointer();
}

Group::pointer example_line_join( Cairo::LineJoin line_join )
{
  Polyline::pointer line = Polyline::create();
  line->add_vertex(-20, 15);
  line->add_vertex(0, -10);
  line->add_vertex(20, 15);
  line->line_style().set_pattern( Cairo::SolidPattern::create_rgb( 0.0, 0.0, 0.0 ) );
  line->line_style().set_join(line_join);
  line->line_style().set_width(25);

  Circle::pointer circle = Circle::create(3);
  circle->set_fill( Cairo::SolidPattern::create_rgb( 0.0, 1.0, 1.0 ) );
  circle->set_translate(0, -10);

  Polyline::pointer inside_line = Polyline::create();
  inside_line->add_vertex(-20, 15);
  inside_line->add_vertex(0, -10);
  inside_line->add_vertex(20, 15);
  inside_line->line_style().set_pattern( Cairo::SolidPattern::create_rgb( 0.0, 1.0, 1.0 ) );
  inside_line->line_style().set_cap(Cairo::LINE_CAP_BUTT);
  inside_line->line_style().set_width(1);

  Group::pointer group = Group::create();
  group->add(line);
  group->add(circle);
  group->add(inside_line);
  return group;
}

Group::pointer example_line_cap( Cairo::LineCap line_cap )
{
  Polyline::pointer inside_line = Polyline::create();
  inside_line->add_vertex(-15, 0);
  inside_line->add_vertex(15, 0);
  inside_line->line_style().set_pattern( Cairo::SolidPattern::create_rgb( 0.0, 1.0, 1.0 ) );
  inside_line->line_style().set_cap(Cairo::LINE_CAP_BUTT);
  inside_line->line_style().set_width(1);

  Circle::pointer left_circle = Circle::create(3);
  left_circle->set_fill( Cairo::SolidPattern::create_rgb( 0.0, 1.0, 1.0 ) );
  left_circle->set_translate(-15, 0);

  Circle::pointer right_circle = Circle::create(3);
  right_circle->set_fill( Cairo::SolidPattern::create_rgb( 0.0, 1.0, 1.0 ) );
  right_circle->set_translate(15, 0);

  Polyline::pointer line = Polyline::create();
  line->add_vertex(-15, 0);
  line->add_vertex(15, 0);
  line->line_style().set_pattern( Cairo::SolidPattern::create_rgb( 0.0, 0.0, 0.0 ) );
  line->line_style().set_cap(line_cap);
  line->line_style().set_width(20);

  Group::pointer group = Group::create();
  group->add(line);
  group->add(left_circle);
  group->add(right_circle);
  group->add(inside_line);
  return group;
}

Polyline::pointer  example_marker( Papyrus::Marker::STYLE style, Papyrus::Marker::FACING facing, bool fill, bool outline) {
  Polyline::pointer line = Polyline::create();
  line->add_vertex(-17, 0);
  line->add_vertex( 17, 0);
  line->line_style().set_pattern( Cairo::SolidPattern::create_rgb( 0.0, 0.0, 0.0 ) );
  line->line_style().set_width(2);

  Marker::pointer marker = Marker::create(style, 15, 15, facing);
  if (fill)
    marker->set_fill( Cairo::SolidPattern::create_rgb( 1.0, 1.0, 1.0 ) );
  if (outline)
    marker->set_outline( Cairo::SolidPattern::create_rgb( 0.0, 0.0, 0.0 ) );
  marker->line_style().set_width(2);
  if (facing == Papyrus::Marker::LEFT)
    line->set_start_marker( marker );
  else
    line->set_end_marker( marker );

  return line;
}

}
